@@ -27,8 +27,12 @@ import threading
 import time
 import traceback
 import json
+from datetime import datetime
 
-import numpy as np
+try:
+    import numpy as np
+except ImportError:
+    np = None
 import pytest
 import pyarrow as pa
 
@@ -149,8 +153,7 @@ class ConstantFlightServer(FlightServerBase):
             yield flight.FlightInfo(
                 pa.schema([]),
                 flight.FlightDescriptor.for_path('/foo'),
-                [],
-                -1, -1
+                []
             )
 
     def do_get(self, context, ticket):
@@ -180,20 +183,16 @@ class MetadataFlightServer(FlightServerBase):
     def do_put(self, context, descriptor, reader, writer):
         counter = 0
         expected_data = [-10, -5, 0, 5, 10]
-        while True:
-            try:
-                batch, buf = reader.read_chunk()
-                assert batch.equals(pa.RecordBatch.from_arrays(
-                    [pa.array([expected_data[counter]])],
-                    ['a']
-                ))
-                assert buf is not None
-                client_counter, = struct.unpack('<i', buf.to_pybytes())
-                assert counter == client_counter
-                writer.write(struct.pack('<i', counter))
-                counter += 1
-            except StopIteration:
-                return
+        for batch, buf in reader:
+            assert batch.equals(pa.RecordBatch.from_arrays(
+                [pa.array([expected_data[counter]])],
+                ['a']
+            ))
+            assert buf is not None
+            client_counter, = struct.unpack('<i', buf.to_pybytes())
+            assert counter == client_counter
+            writer.write(struct.pack('<i', counter))
+            counter += 1
 
     @staticmethod
     def number_batches(table):
@@ -252,10 +251,14 @@ class GetInfoFlightServer(FlightServerBase):
                 flight.FlightEndpoint(
                     b'',
                     [flight.Location.for_grpc_tcp('localhost', 5005)],
+                    pa.scalar("2023-04-05T12:34:56.789012345").cast(pa.timestamp("ns")),
+                    "endpoint app metadata"
                 ),
             ],
-            -1,
-            -1,
+            1,
+            42,
+            True,
+            "info app metadata"
         )
 
     def get_schema(self, context, descriptor):
@@ -388,8 +391,7 @@ class ErrorFlightServer(FlightServerBase):
         yield flight.FlightInfo(
             pa.schema([]),
             flight.FlightDescriptor.for_path('/foo'),
-            [],
-            -1, -1
+            []
         )
         raise flight.FlightInternalError("foo")
 
@@ -577,7 +579,7 @@ class NoopAuthHandler(ServerAuthHandler):
 
 def case_insensitive_header_lookup(headers, lookup_key):
     """Lookup the value of given key in the given headers.
-       The key lookup is case insensitive.
+       The key lookup is case-insensitive.
     """
     for key in headers:
         if key.lower() == lookup_key.lower():
@@ -833,7 +835,7 @@ class MultiHeaderClientMiddleware(ClientMiddleware):
     def received_headers(self, headers):
         # Let the test code know what the last set of headers we
         # received were.
-        self.factory.last_headers = headers
+        self.factory.last_headers.update(headers)
 
 
 class MultiHeaderServerMiddlewareFactory(ServerMiddlewareFactory):
@@ -877,14 +879,18 @@ def test_repr():
     descriptor_repr = "<pyarrow.flight.FlightDescriptor cmd=b'foo'>"
     endpoint_repr = ("<pyarrow.flight.FlightEndpoint "
                      "ticket=<pyarrow.flight.Ticket ticket=b'foo'> "
-                     "locations=[]>")
+                     "locations=[] "
+                     "expiration_time=2023-04-05 12:34:56+00:00 "
+                     "app_metadata=b'endpoint app metadata'>")
     info_repr = (
         "<pyarrow.flight.FlightInfo "
         "schema= "
         "descriptor=<pyarrow.flight.FlightDescriptor path=[]> "
         "endpoints=[] "
-        "total_records=-1 "
-        "total_bytes=-1>")
+        "total_records=1 "
+        "total_bytes=42 "
+        "ordered=True "
+        "app_metadata=b'test app metadata'>")
     location_repr = "<pyarrow.flight.Location b'grpc+tcp://localhost:1234'>"
     result_repr = "<pyarrow.flight.Result body=(3 bytes)>"
     schema_result_repr = "<pyarrow.flight.SchemaResult schema=()>"
@@ -894,9 +900,15 @@ def test_repr():
     assert repr(flight.ActionType("foo", "bar")) == action_type_repr
     assert repr(flight.BasicAuth("user", "pass")) == basic_auth_repr
     assert repr(flight.FlightDescriptor.for_command("foo")) == descriptor_repr
-    assert repr(flight.FlightEndpoint(b"foo", [])) == endpoint_repr
+    endpoint = flight.FlightEndpoint(
+        b"foo", [], pa.scalar("2023-04-05T12:34:56").cast(pa.timestamp("s")),
+        b"endpoint app metadata"
+    )
+    assert repr(endpoint) == endpoint_repr
     info = flight.FlightInfo(
-        pa.schema([]), flight.FlightDescriptor.for_path(), [], -1, -1)
+        pa.schema([]), flight.FlightDescriptor.for_path(), [],
+        1, 42, True, b"test app metadata"
+    )
     assert repr(info) == info_repr
     assert repr(flight.Location("grpc+tcp://localhost:1234")) == location_repr
     assert repr(flight.Result(b"foo")) == result_repr
@@ -908,25 +920,97 @@ def test_repr():
     with pytest.raises(TypeError):
         flight.Action("foo", None)
 
+    with pytest.raises(TypeError):
+        flight.FlightEndpoint(object(), [])
+    with pytest.raises(TypeError):
+        flight.FlightEndpoint("foo", ["grpc://test", b"grpc://test", object()])
+    with pytest.raises(TypeError):
+        flight.FlightEndpoint("foo", [], expiration_time="2023-04-05T01:02:03")
+    with pytest.raises(TypeError):
+        flight.FlightEndpoint("foo", [], expiration_time=datetime(2023, 4, 5, 1, 2, 3))
+    with pytest.raises(TypeError):
+        flight.FlightEndpoint("foo", [], app_metadata=object())
+
 
 def test_eq():
     items = [
+        lambda: (flight.Action("foo", b""), flight.Action("bar", b"")),
         lambda: (flight.Action("foo", b""), flight.Action("foo", b"bar")),
         lambda: (flight.ActionType("foo", "bar"),
                  flight.ActionType("foo", "baz")),
         lambda: (flight.BasicAuth("user", "pass"),
                  flight.BasicAuth("user2", "pass")),
+        lambda: (flight.BasicAuth("user", "pass"),
+                 flight.BasicAuth("user", "pass2")),
         lambda: (flight.FlightDescriptor.for_command("foo"),
                  flight.FlightDescriptor.for_path("foo")),
         lambda: (flight.FlightEndpoint(b"foo", []),
-                 flight.FlightEndpoint(b"", [])),
+                 flight.FlightEndpoint(b"bar", [])),
+        lambda: (
+            flight.FlightEndpoint(
+                b"foo", [flight.Location("grpc+tcp://localhost:1234")]),
+            flight.FlightEndpoint(
+                b"foo", [flight.Location("grpc+tls://localhost:1234")])
+        ),
+        lambda: (
+            flight.FlightEndpoint(
+                b"foo", [], pa.scalar("2023-04-05T12:34:56").cast(pa.timestamp("s"))),
+            flight.FlightEndpoint(
+                b"foo", [],
+                pa.scalar("2023-04-05T12:34:56.789").cast(pa.timestamp("ms")))),
+        lambda: (flight.FlightEndpoint(b"foo", [], app_metadata=b''),
+                 flight.FlightEndpoint(b"foo", [], app_metadata=b'meta')),
         lambda: (
             flight.FlightInfo(
                 pa.schema([]),
-                flight.FlightDescriptor.for_path(), [], -1, -1),
+                flight.FlightDescriptor.for_path(), []),
+            flight.FlightInfo(
+                pa.schema([("ints", pa.int64())]),
+                flight.FlightDescriptor.for_path(), [])),
+        lambda: (
             flight.FlightInfo(
                 pa.schema([]),
-                flight.FlightDescriptor.for_command(b"foo"), [], -1, 42)),
+                flight.FlightDescriptor.for_path(), []),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_command(b"foo"), [])),
+        lambda: (
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(),
+                [flight.FlightEndpoint(b"foo", [])]),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(),
+                [flight.FlightEndpoint(b"bar", [])])),
+        lambda: (
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], total_records=-1),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], total_records=1)),
+        lambda: (
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], total_bytes=-1),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], total_bytes=42)),
+        lambda: (
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], ordered=False),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], ordered=True)),
+        lambda: (
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], app_metadata=b""),
+            flight.FlightInfo(
+                pa.schema([]),
+                flight.FlightDescriptor.for_path(), [], app_metadata=b"meta")),
         lambda: (flight.Location("grpc+tcp://localhost:1234"),
                  flight.Location("grpc+tls://localhost:1234")),
         lambda: (flight.Result(b"foo"), flight.Result(b"bar")),
@@ -938,11 +1022,33 @@ def test_eq():
     for gen in items:
         lhs1, rhs1 = gen()
         lhs2, rhs2 = gen()
+        assert lhs1 == lhs1
         assert lhs1 == lhs2
+        assert lhs2 == lhs1
+        assert rhs1 == rhs1
         assert rhs1 == rhs2
+        assert rhs2 == rhs1
         assert lhs1 != rhs1
 
 
+def test_flight_info_defaults():
+    fi1 = flight.FlightInfo(pa.schema([]), flight.FlightDescriptor.for_path(), [])
+    fi2 = flight.FlightInfo(
+        pa.schema([]),
+        flight.FlightDescriptor.for_path(), [], total_records=-1, total_bytes=-1)
+    fi3 = flight.FlightInfo(
+        pa.schema([]),
+        flight.FlightDescriptor.for_path(), [], total_records=None, total_bytes=None)
+
+    assert fi1.total_records == -1
+    assert fi2.total_records == -1
+    assert fi3.total_records == -1
+
+    assert fi1.total_bytes == -1
+    assert fi2.total_bytes == -1
+    assert fi3.total_bytes == -1
+
+
 def test_flight_server_location_argument():
     locations = [
         None,
@@ -960,6 +1066,7 @@ def test_server_exit_reraises_exception():
             raise ValueError()
 
 
+@pytest.mark.threading
 @pytest.mark.slow
 def test_client_wait_for_available():
     location = ('localhost', find_free_port())
@@ -1062,12 +1169,20 @@ def test_flight_get_info():
     with GetInfoFlightServer() as server:
         client = FlightClient(('localhost', server.port))
         info = client.get_flight_info(flight.FlightDescriptor.for_command(b''))
-        assert info.total_records == -1
-        assert info.total_bytes == -1
+        assert info.total_records == 1
+        assert info.total_bytes == 42
+        assert info.ordered
+        assert info.app_metadata == b"info app metadata"
         assert info.schema == pa.schema([('a', pa.int32())])
         assert len(info.endpoints) == 2
         assert len(info.endpoints[0].locations) == 1
+        assert info.endpoints[0].expiration_time is None
+        assert info.endpoints[0].app_metadata == b""
         assert info.endpoints[0].locations[0] == flight.Location('grpc://test')
+        assert info.endpoints[1].expiration_time == \
+            pa.scalar("2023-04-05T12:34:56.789012345+00:00") \
+              .cast(pa.timestamp("ns", "UTC"))
+        assert info.endpoints[1].app_metadata == b"endpoint app metadata"
         assert info.endpoints[1].locations[0] == \
             flight.Location.for_grpc_tcp('localhost', 5005)
 
@@ -1495,7 +1610,7 @@ def test_tls_override_hostname():
     """Check that incorrectly overriding the hostname fails."""
     certs = example_tls_certs()
 
-    with ConstantFlightServer(tls_certificates=certs["certificates"]) as s,\
+    with ConstantFlightServer(tls_certificates=certs["certificates"]) as s, \
         flight.connect(('localhost', s.port),
                        tls_root_certs=certs["root_cert"],
                        override_hostname="fakehostname") as client:
@@ -1511,6 +1626,24 @@ def test_flight_do_get_metadata():
     table = pa.Table.from_arrays(data, names=['a'])
 
     batches = []
+    with MetadataFlightServer() as server, \
+            FlightClient(('localhost', server.port)) as client:
+        reader = client.do_get(flight.Ticket(b''))
+        idx = 0
+        for batch, metadata in reader:
+            batches.append(batch)
+            server_idx, = struct.unpack('<i', metadata.to_pybytes())
+            assert idx == server_idx
+            idx += 1
+        data = pa.Table.from_batches(batches)
+        assert data.equals(table)
+
+
+def test_flight_metadata_record_batch_reader_iterator():
+    """Verify the iterator interface works as expected."""
+    batches1 = []
+    batches2 = []
+
     with MetadataFlightServer() as server, \
             FlightClient(('localhost', server.port)) as client:
         reader = client.do_get(flight.Ticket(b''))
@@ -1518,14 +1651,24 @@ def test_flight_do_get_metadata():
         while True:
             try:
                 batch, metadata = reader.read_chunk()
-                batches.append(batch)
+                batches1.append(batch)
                 server_idx, = struct.unpack('<i', metadata.to_pybytes())
                 assert idx == server_idx
                 idx += 1
             except StopIteration:
                 break
-        data = pa.Table.from_batches(batches)
-        assert data.equals(table)
+
+    with MetadataFlightServer() as server, \
+            FlightClient(('localhost', server.port)) as client:
+        reader = client.do_get(flight.Ticket(b''))
+        idx = 0
+        for batch, metadata in reader:
+            batches2.append(batch)
+            server_idx, = struct.unpack('<i', metadata.to_pybytes())
+            assert idx == server_idx
+            idx += 1
+
+    assert batches1 == batches2
 
 
 def test_flight_do_get_metadata_v4():
@@ -1563,6 +1706,7 @@ def test_flight_do_put_metadata():
                 assert idx == server_idx
 
 
+@pytest.mark.numpy
 def test_flight_do_put_limit():
     """Try a simple do_put call with a size limit."""
     large_batch = pa.RecordBatch.from_arrays([
@@ -1603,6 +1747,7 @@ def test_cancel_do_get():
             reader.read_chunk()
 
 
+@pytest.mark.threading
 @pytest.mark.slow
 def test_cancel_do_get_threaded():
     """Test canceling a DoGet operation from another thread."""
@@ -1677,21 +1822,29 @@ def test_roundtrip_types():
             flight.FlightEndpoint(
                 b'',
                 [flight.Location.for_grpc_tcp('localhost', 5005)],
+                pa.scalar("2023-04-05T12:34:56.789012345").cast(pa.timestamp("ns")),
+                b'endpoint app metadata'
             ),
         ],
-        -1,
-        -1,
+        1,
+        42,
+        True,
+        b'test app metadata'
     )
     info2 = flight.FlightInfo.deserialize(info.serialize())
     assert info.schema == info2.schema
     assert info.descriptor == info2.descriptor
     assert info.total_bytes == info2.total_bytes
     assert info.total_records == info2.total_records
+    assert info.ordered == info2.ordered
+    assert info.app_metadata == info2.app_metadata
     assert info.endpoints == info2.endpoints
 
     endpoint = flight.FlightEndpoint(
         ticket,
-        ['grpc://test', flight.Location.for_grpc_tcp('localhost', 5005)]
+        ['grpc://test', flight.Location.for_grpc_tcp('localhost', 5005)],
+        pa.scalar("2023-04-05T12:34:56").cast(pa.timestamp("s")),
+        b'endpoint app metadata'
     )
     assert endpoint == flight.FlightEndpoint.deserialize(endpoint.serialize())
 
@@ -2067,15 +2220,14 @@ class CancelFlightServer(FlightServerBase):
             time.sleep(0.5)
 
 
+@pytest.mark.threading
 def test_interrupt():
     if threading.current_thread().ident != threading.main_thread().ident:
         pytest.skip("test only works from main Python thread")
-    # Skips test if not available
-    raise_signal = util.get_raise_signal()
 
     def signal_from_thread():
         time.sleep(0.5)
-        raise_signal(signal.SIGINT)
+        signal.raise_signal(signal.SIGINT)
 
     exc_types = (KeyboardInterrupt, pa.ArrowCancelled)
 
@@ -2323,3 +2475,44 @@ def test_do_put_does_not_crash_when_schema_is_none():
     with pytest.raises(TypeError, match=msg):
         client.do_put(flight.FlightDescriptor.for_command('foo'),
                       schema=None)
+
+
+def test_headers_trailers():
+    """Ensure that server-sent headers/trailers make it through."""
+
+    class HeadersTrailersFlightServer(FlightServerBase):
+        def get_flight_info(self, context, descriptor):
+            context.add_header("x-header", "header-value")
+            context.add_header("x-header-bin", "header\x01value")
+            context.add_trailer("x-trailer", "trailer-value")
+            context.add_trailer("x-trailer-bin", "trailer\x01value")
+            return flight.FlightInfo(
+                pa.schema([]),
+                descriptor,
+                []
+            )
+
+    class HeadersTrailersMiddlewareFactory(ClientMiddlewareFactory):
+        def __init__(self):
+            self.headers = []
+
+        def start_call(self, info):
+            return HeadersTrailersMiddleware(self)
+
+    class HeadersTrailersMiddleware(ClientMiddleware):
+        def __init__(self, factory):
+            self.factory = factory
+
+        def received_headers(self, headers):
+            for key, values in headers.items():
+                for value in values:
+                    self.factory.headers.append((key, value))
+
+    factory = HeadersTrailersMiddlewareFactory()
+    with HeadersTrailersFlightServer() as server, \
+            FlightClient(("localhost", server.port), middleware=[factory]) as client:
+        client.get_flight_info(flight.FlightDescriptor.for_path(""))
+        assert ("x-header", "header-value") in factory.headers
+        assert ("x-header-bin", b"header\x01value") in factory.headers
+        assert ("x-trailer", "trailer-value") in factory.headers
+        assert ("x-trailer-bin", b"trailer\x01value") in factory.headers
